@@ -39,7 +39,49 @@
 
 /* === C L A S S E S & S T R U C T S ======================================= */
 
-_ETL_BEGIN_NAMESPACE
+namespace etl {
+
+class clamping
+{
+public:
+	typedef bool func(int&, int);
+
+	inline static bool lock(int&, int)
+		{ return false; }
+
+	inline static bool pass(int&, int)
+		{ return true; }
+
+	inline static bool truncate(int &x, int bound)
+		{ return x >= 0 && x < bound; }
+
+	inline static bool clamp(int &x, int bound) {
+		if (bound <= 0) return false;
+		if (x < 0) x = 0; else
+			if (x >= bound) x = bound - 1;
+		return true;
+	}
+
+	inline static int repeat(int &x, int bound) {
+		if (bound <= 0) return false;
+		x %= bound;
+		if (x < 0) x += bound;
+		return true;
+	}
+
+	inline static int mirror(int &x, int bound) {
+		if (bound <= 0) return false;
+		x = abs(x);
+		return x < bound;
+	}
+
+	inline static int mirror_repeat(int &x, int bound) {
+		if (bound <= 0) return false;
+		x = abs((abs(x) + bound)%(2*bound) - bound);
+		return true;
+	}
+};
+
 
 template <typename T, typename AT>
 class value_prep
@@ -60,20 +102,21 @@ public:
 	typedef CT coord_type;
 	typedef ST source_type;
 	typedef coord_type float_type;
+	typedef value_type func(const void*, const coord_type, const coord_type);
 
-	inline static void prepare_coord(int w, const coord_type x, int &u, float_type &a) {
-		if (x<0) { u=0; a=0; }
-		else if (x>w-1.00001) { u=w-2; a=1; }
-		else {
-			u=floor_to_int(x);
-			a=float_type(x)-float_type(u);
-		}
+	template<typename T, T wrap_func(const VT&), func sampler_func>
+	inline static T wrap(const void *surface, const coord_type x, const coord_type y)
+		{ return wrap_func(sampler_func(surface, x, y)); }
+
+	inline static void prepare_coord(const coord_type x, int &u, float_type &a) {
+		u=floor_to_int(x);
+		a=float_type(x)-float_type(u);
 	}
 
-	inline static void prepare_coords(int w, int h, const coord_type x, const coord_type y, int &u, int &v, float_type &a, float_type &b)
+	inline static void prepare_coords(const coord_type x, const coord_type y, int &u, int &v, float_type &a, float_type &b)
 	{
-		prepare_coord(w, x, u, a);
-		prepare_coord(h, y, v, b);
+		prepare_coord(x, u, a);
+		prepare_coord(y, v, b);
 	}
 
 	inline static void fill_cubic_polinomial(float_type x, float_type tx[])
@@ -84,11 +127,15 @@ public:
 		tx[3] = float_type(0.5)*x*x*(x-float_type(1));						                // -t^2 + t^3
 	}
 
+	//! Nearest sample
+	static value_type nearest_sample(const void *surface, const coord_type x, const coord_type y)
+		{ return (value_type)reader(surface, round_to_int(x), round_to_int(y)); }
+
 	//! Linear sample
-	static value_type linear_sample(const void *surface, int w, int h, const coord_type x, const coord_type y)
+	static value_type linear_sample(const void *surface, const coord_type x, const coord_type y)
 	{
 		int u, v; float_type a, b;
-		prepare_coords(w, h, x, y, u, v, a, b);
+		prepare_coords(x, y, u, v, a, b);
 
 		const float_type c(float_type(1)-a), d(float_type(1)-b);
 
@@ -99,10 +146,10 @@ public:
 	}
 
 	//! Cosine sample
-	static value_type cosine_sample(const void *surface, int w, int h, const coord_type x, const coord_type y)
+	static value_type cosine_sample(const void *surface, const coord_type x, const coord_type y)
 	{
 		int u, v; float_type a, b;
-		prepare_coords(w, h, x, y, u, v, a, b);
+		prepare_coords(x, y, u, v, a, b);
 
 		a=(float_type(1) - cos(a*float_type(3.1415927)))*float_type(0.5);
 		b=(float_type(1) - cos(b*float_type(3.1415927)))*float_type(0.5);
@@ -116,7 +163,7 @@ public:
 	}
 
 	//! Cubic sample
-	static value_type cubic_sample(const void *surface, int w, int h, const coord_type x, const coord_type y)
+	static value_type cubic_sample(const void *surface, const coord_type x, const coord_type y)
 	{
 		//Using catmull rom interpolation because it doesn't blur at all
 		//bezier curve with intermediate ctrl pts: 0.5/3(p(i+1) - p(i-1)) and similar
@@ -127,14 +174,6 @@ public:
 		int xa[] = { xi-1, xi, xi+1, xi+2 };
 		int ya[] = { yi-1, yi, yi+1, yi+2 };
 
-		// clamp all
-		const int ww = w - 1;
-		const int hh = h - 1;
-		xa[0] < 0  && ((xa[1] < 0  && ((xa[2] < 0  && ((xa[3] < 0  && (xa[3] = 0 )), (xa[2] = 0 ))), (xa[1] = 0 ))), (xa[0] = 0 ));
-		ya[0] < 0  && ((ya[1] < 0  && ((ya[2] < 0  && ((ya[3] < 0  && (ya[3] = 0 )), (ya[2] = 0 ))), (ya[1] = 0 ))), (ya[0] = 0 ));
-		xa[3] > ww && ((xa[2] > ww && ((xa[1] > ww && ((xa[0] > ww && (xa[0] = ww)), (xa[1] = ww))), (xa[2] = ww))), (xa[3] = ww));
-		ya[3] > hh && ((ya[2] > hh && ((ya[1] > hh && ((ya[0] > hh && (ya[0] = hh)), (ya[1] = hh))), (ya[2] = hh))), (ya[3] = hh));
-
 		// offset
 		const float_type xf = float_type(x)-float_type(xi);
 		const float_type yf = float_type(y)-float_type(yi);
@@ -267,7 +306,7 @@ public:
 		if(s.data_)
 		{
 			assert(data_);
-			memcpy(data_,s.data_,abs(pitch_)*h_);
+			memcpy(data_,s.data_,std::abs(pitch_)*h_);
 		}
 	}
 
@@ -442,7 +481,7 @@ public:
 	{
 		assert(data_);
 		if(pitch_==(signed int)sizeof(value_type)*w_)
-			memset(data_,0,h_*pitch_);
+			memset(static_cast<void*>(data_), 0, h_*pitch_);
 		else
 			fill(value_type());
 	}
@@ -485,346 +524,60 @@ public:
 	const_pen get_pen(int x, int y)const { assert(data_); return begin().move(x,y); }
 	const_pen end()const { assert(data_); return get_pen(w_,h_); }
 
-	inline static value_type reader(const void *surf, int x, int y)
-		{ return (*(const surface*)surf)[y][x]; }
-	inline static accumulator_type reader_cook(const void *surf, int x, int y)
-		{ return (*(const surface*)surf).cooker_.cook(reader(surf, x, y)); }
+	template< clamping::func clamp_x = clamping::clamp,
+			  clamping::func clamp_y = clamping::clamp >
+	inline static value_type reader(const void *surf, int x, int y) {
+		const surface &s = *(const surface*)surf;
+		return clamp_x(x, s.get_w()) && clamp_y(y, s.get_h()) ? s[y][x] : value_type();
+	}
 
-	typedef sampler<accumulator_type, float, accumulator_type, surface::reader_cook> sampler_cook;
-	typedef sampler<accumulator_type, float, value_type, surface::reader> sampler_nocook;
+	template< clamping::func clamp_x = clamping::clamp,
+			  clamping::func clamp_y = clamping::clamp >
+	inline static accumulator_type reader_cook(const void *surf, int x, int y) {
+		const surface &s = *(const surface*)surf;
+		return clamp_x(x, s.get_w()) && clamp_y(y, s.get_h()) ? s.cooker_.cook(s[y][x]) : value_type();
+	}
+
+	template<typename ReaderType, ReaderType reader(const void*, int, int)>
+	class sampler: public etl::sampler<accumulator_type, float, ReaderType, reader> { };
+
+	typedef sampler<accumulator_type, surface::reader_cook> sampler_cook;
+	typedef sampler<value_type, surface::reader> sampler_nocook;
+
+	//! Nearest sample
+	value_type nearest_sample(const float x, const float y)const
+		{ return cooker_.uncook(sampler_cook::nearest_sample(this, x, y)); }
+
+	//! Nearest sample for already "cooked" surfaces
+	value_type nearest_sample_cooked(const float x, const float y)const
+		{ return (value_type)(sampler_nocook::nearest_sample(this, x, y)); }
 
 	//! Linear sample
 	value_type linear_sample(const float x, const float y)const
-		{ return cooker_.uncook(sampler_cook::linear_sample(this, w_, h_, x, y)); }
+		{ return cooker_.uncook(sampler_cook::linear_sample(this, x, y)); }
 
 	//! Linear sample for already "cooked" surfaces
 	value_type linear_sample_cooked(const float x, const float y)const
-		{ return (value_type)(sampler_nocook::linear_sample(this, w_, h_, x, y)); }
+		{ return (value_type)(sampler_nocook::linear_sample(this, x, y)); }
 
 	//! Cosine sample
 	value_type cosine_sample(const float x, const float y)const
-		{ return cooker_.uncook(sampler_cook::cosine_sample(this, w_, h_, x, y)); }
+		{ return cooker_.uncook(sampler_cook::cosine_sample(this, x, y)); }
 
 	//! Cosine sample for already "cooked" surfaces
 	value_type cosine_sample_cooked(const float x, const float y)const
-		{ return (value_type)(sampler_nocook::cosine_sample(this, w_, h_, x, y)); }
+		{ return (value_type)(sampler_nocook::cosine_sample(this, x, y)); }
 
 	//! Cubic sample
 	value_type cubic_sample(float x, float y)const
-		{ return cooker_.uncook(sampler_cook::cubic_sample(this, w_, h_, x, y)); }
+		{ return cooker_.uncook(sampler_cook::cubic_sample(this, x, y)); }
 
 	//! Cubic sample for already "cooked" surfaces
 	value_type cubic_sample_cooked(float x, float y)const
-		{ return (value_type)(sampler_nocook::cubic_sample(this, w_, h_, x, y)); }
-
-	//! Rectangle sample
-	value_type	sample_rect(float x0,float y0,float x1,float y1) const
-	{
-		const surface &s = *this;
-
-		//assumes it's clamped to the boundary of the image
-		//force min max relationship for x0,x1 and y0,y1
-		if(x0 > x1) std::swap(x0,x1);
-		if(y0 > y1) std::swap(y0,y1);
-
-		//local variable madness
-		//all things that want to inter-operate should provide a default value constructor for = 0
-		accumulator_type acum = 0;
-		int xi=0,yi=0;
-
-		int	xib=(int)floor(x0),
-			xie=(int)floor(x1);
-
-		int	yib=(int)floor(y0),
-			yie=(int)floor(y1);
-
-		//the weight for the pixel should remain the same...
-		float weight = (y1-y0)*(x1-x0);
-		assert(weight != 0);
-
-		float ylast = y0, xlastb = x0;
-		const_pen	pen_ = s.get_pen(xib,yib);
-
-		for(yi = yib; yi < yie; ylast = ++yi, pen_.inc_y())
-		{
-			const float yweight = yi+1 - ylast;
-
-			float xlast = xlastb;
-			for(xi = xib; xi < xie; xlast = ++xi, pen_.inc_x())
-			{
-				const float w = yweight*(xi+1 - xlast);
-				acum += cooker_.cook(pen_.get_value())*w;
-			}
-
-			//post... with next being fractional...
-			const float w = yweight*(x1 - xlast);
-			acum += cooker_.cook(pen_.get_value())*w;
-
-			pen_.dec_x(xie-xib);
-		}
-
-		//post in y direction... must have all x...
-		{
-			const float yweight = y1 - ylast;
-
-			float xlast = xlastb;
-			for(xi = xib; xi < xie; xlast = ++xi)
-			{
-				const float w = yweight*(xi+1 - xlast);
-
-				acum += cooker_.cook(pen_.get_value())*w;
-			}
-
-			//post... with next being fractional...
-			const float w = yweight*(x1 - xlast);
-			acum += cooker_.cook(pen_.get_value())*w;
-		}
-
-		acum *= 1/weight;
-		return cooker_.uncook(acum);
-	}
-
-	//! Rectangle sample for already "cooked" surfaces
-	value_type	sample_rect_cooked(float x0,float y0,float x1,float y1) const
-	{
-		const surface &s = *this;
-		
-		//assumes it's clamped to the boundary of the image
-		//force min max relationship for x0,x1 and y0,y1
-		if(x0 > x1) std::swap(x0,x1);
-		if(y0 > y1) std::swap(y0,y1);
-		
-		//local variable madness
-		//all things that want to inter-operate should provide a default value constructor for = 0
-		accumulator_type acum = 0;
-		int xi=0,yi=0;
-		
-		int	xib=(int)floor(x0),
-		xie=(int)floor(x1);
-		
-		int	yib=(int)floor(y0),
-		yie=(int)floor(y1);
-		
-		//the weight for the pixel should remain the same...
-		float weight = (y1-y0)*(x1-x0);
-		assert(weight != 0);
-		
-		float ylast = y0, xlastb = x0;
-		const_pen	pen_ = s.get_pen(xib,yib);
-		
-		for(yi = yib; yi < yie; ylast = ++yi, pen_.inc_y())
-		{
-			const float yweight = yi+1 - ylast;
-			
-			float xlast = xlastb;
-			for(xi = xib; xi < xie; xlast = ++xi, pen_.inc_x())
-			{
-				const float w = yweight*(xi+1 - xlast);
-				acum += pen_.get_value()*w;
-			}
-			
-			//post... with next being fractional...
-			const float w = yweight*(x1 - xlast);
-			acum += pen_.get_value()*w;
-			
-			pen_.dec_x(xie-xib);
-		}
-		
-		//post in y direction... must have all x...
-		{
-			const float yweight = y1 - ylast;
-			
-			float xlast = xlastb;
-			for(xi = xib; xi < xie; xlast = ++xi)
-			{
-				const float w = yweight*(xi+1 - xlast);
-				
-				acum += pen_.get_value()*w;
-			}
-			
-			//post... with next being fractional...
-			const float w = yweight*(x1 - xlast);
-			acum += pen_.get_value()*w;
-		}
-		
-		acum *= 1/weight;
-		return (value_type)(acum);
-	}
-
-	//! Rectangle sample clipped
-	value_type	sample_rect_clip(float x0,float y0,float x1,float y1) const
-	{
-		const surface &s = *this;
-
-		//assumes it's clamped to the boundary of the image
-		//force min max relationship for x0,x1 and y0,y1
-		if(x0 > x1) std::swap(x0,x1);
-		if(y0 > y1) std::swap(y0,y1);
-
-		//local variable madness
-		//all things that want to inter-operate should provide a default value constructor for = 0
-		accumulator_type acum = accumulator_type();
-		int xi=0,yi=0;
-
-		int	xib=(int)floor(x0),
-			xie=(int)floor(x1);
-
-		int	yib=(int)floor(y0),
-			yie=(int)floor(y1);
-
-		//the weight for the pixel should remain the same...
-		float weight = (y1-y0)*(x1-x0);
-
-		assert(weight != 0);
-
-		//clip to the input region
-		if(x0 >= s.get_w() || x1 <= 0) return acum;
-		if(y0 >= s.get_h() || y1 <= 0) return acum;
-
-		if(x0 < 0) { x0 = 0; xib = 0; }
-		if(x1 >= s.get_w())
-		{
-			x1 = s.get_w(); //want to be just below the last pixel...
-			xie = s.get_w()-1;
-		}
-
-		if(y0 < 0) { y0 = 0; yib = 0; }
-		if(y1 >= s.get_h())
-		{
-			y1 = s.get_h(); //want to be just below the last pixel...
-			yie = s.get_h()-1;
-		}
-
-		float ylast = y0, xlastb = x0;
-		const_pen	pen = s.get_pen(xib,yib);
-
-		for(yi = yib; yi < yie; ylast = ++yi, pen.inc_y())
-		{
-			const float yweight = yi+1 - ylast;
-
-			float xlast = xlastb;
-			for(xi = xib; xi < xie; xlast = ++xi, pen.inc_x())
-			{
-				const float w = yweight*(xi+1 - xlast);
-				acum += cooker_.cook(pen.get_value())*w;
-			}
-
-			//post... with next being fractional...
-			const float w = yweight*(x1 - xlast);
-			acum += cooker_.cook(pen.get_value())*w;
-
-			pen.dec_x(xie-xib);
-		}
-
-		//post in y direction... must have all x...
-		{
-			const float yweight = y1 - ylast;
-
-			float xlast = xlastb;
-			for(xi = xib; xi < xie; xlast = ++xi)
-			{
-				const float w = yweight*(xi+1 - xlast);
-
-				acum += cooker_.cook(pen.get_value())*w;
-			}
-
-			//post... with next being fractional...
-			const float w = yweight*(x1 - xlast);
-			acum += cooker_.cook(pen.get_value())*w;
-		}
-
-		acum *= 1/weight;
-		return cooker_.uncook(acum);
-	}
-
-	//! Rectangle sample clipped for already "cooked" surfaces
-	value_type	sample_rect_clip_cooked(float x0,float y0,float x1,float y1) const
-	{
-		const surface &s = *this;
-		
-		//assumes it's clamped to the boundary of the image
-		//force min max relationship for x0,x1 and y0,y1
-		if(x0 > x1) std::swap(x0,x1);
-		if(y0 > y1) std::swap(y0,y1);
-		
-		//local variable madness
-		//all things that want to inter-operate should provide a default value constructor for = 0
-		accumulator_type acum = 0;
-		int xi=0,yi=0;
-		
-		int	xib=(int)floor(x0),
-		xie=(int)floor(x1);
-		
-		int	yib=(int)floor(y0),
-		yie=(int)floor(y1);
-		
-		//the weight for the pixel should remain the same...
-		float weight = (y1-y0)*(x1-x0);
-		
-		assert(weight != 0);
-		
-		//clip to the input region
-		if(x0 >= s.get_w() || x1 <= 0) return acum;
-		if(y0 >= s.get_h() || y1 <= 0) return acum;
-		
-		if(x0 < 0) { x0 = 0; xib = 0; }
-		if(x1 >= s.get_w())
-		{
-			x1 = s.get_w(); //want to be just below the last pixel...
-			xie = s.get_w()-1;
-		}
-		
-		if(y0 < 0) { y0 = 0; yib = 0; }
-		if(y1 >= s.get_h())
-		{
-			y1 = s.get_h(); //want to be just below the last pixel...
-			yie = s.get_h()-1;
-		}
-		
-		float ylast = y0, xlastb = x0;
-		const_pen	pen = s.get_pen(xib,yib);
-		
-		for(yi = yib; yi < yie; ylast = ++yi, pen.inc_y())
-		{
-			const float yweight = yi+1 - ylast;
-			
-			float xlast = xlastb;
-			for(xi = xib; xi < xie; xlast = ++xi, pen.inc_x())
-			{
-				const float w = yweight*(xi+1 - xlast);
-				acum += pen.get_value()*w;
-			}
-			
-			//post... with next being fractional...
-			const float w = yweight*(x1 - xlast);
-			acum += pen.get_value()*w;
-			
-			pen.dec_x(xie-xib);
-		}
-		
-		//post in y direction... must have all x...
-		{
-			const float yweight = y1 - ylast;
-			
-			float xlast = xlastb;
-			for(xi = xib; xi < xie; xlast = ++xi)
-			{
-				const float w = yweight*(xi+1 - xlast);
-				
-				acum += pen.get_value()*w;
-			}
-			
-			//post... with next being fractional...
-			const float w = yweight*(x1 - xlast);
-			acum += pen.get_value()*w;
-		}
-		
-		acum *= 1/weight;
-		return (value_type)(acum);
-	}
+		{ return (value_type)(sampler_nocook::cubic_sample(this, x, y)); }
 };
 
-_ETL_END_NAMESPACE
+};
 
 /* === T Y P E D E F S ===================================================== */
 
